# MIT License

# Copyright (c) 2016 Diogo Dutra <dutradda@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from myreco.item_types.indices_map import ItemsIndicesMap
from swaggerit.models.orm.redis_elsearch import ModelRedisElSearchMeta


class _StoreItemsModelBaseMeta(ModelRedisElSearchMeta):

    def __init__(cls, name, bases_classes, attributes):
        super().__init__(name, bases_classes, attributes)
        cls.indices_map = ItemsIndicesMap(cls)

    async def insert(cls, session, objs, skip_validation=False, **kwargs):
        if not skip_validation:
            cls._validate_objs(objs, 'insert')

        return await ModelRedisElSearchMeta.insert(cls, session, objs, **kwargs)

    def _validate_objs(cls, objs, type_):
        validator_name = type_ + '_validator'
        if hasattr(cls, validator_name):
            validator = getattr(cls, validator_name)
            validator.validate(objs)

    async def update(cls, session, objs, ids=None, skip_validation=False, **kwargs):
        if not skip_validation:
            cls._validate_objs(objs, 'update')

        return await ModelRedisElSearchMeta.update(cls, session, objs, **kwargs)

    async def get(cls, session, ids=None, limit=None, offset=None, **kwargs):
        items_per_page, page = kwargs.get('items_per_page', 1000), kwargs.get('page', 1)
        limit = items_per_page * page
        offset = items_per_page * (page-1)
        return await \
            ModelRedisElSearchMeta.get(cls, session, ids=ids, limit=limit, offset=offset, **kwargs)

    async def get_all(cls, session, **kwargs):
        return await ModelRedisElSearchMeta.get(cls, session, **kwargs)

    async def search(cls, session, pattern, page=1, size=100):
        if page < 1:
            return []

        return await ModelRedisElSearchMeta.search(cls, session, pattern, page-1, size)



# MIT License

# Copyright (c) 2016 Diogo Dutra <dutradda@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from myreco.exceptions import ModuleObjectLoaderError
from swaggerit.models.orm.factory import FactoryOrmModels
from swaggerit.utils import get_model, get_swagger_json
from importlib import import_module
from copy import deepcopy
import asyncio
import os
import re


class ModuleObjectLoader(object):
    _objects = dict()

    @classmethod
    def load(cls, config):
        key = '{}.{}'.format(config['path'], config['object_name'])
        obj = cls._objects.get(key)

        if obj is None:
            try:
                module = import_module(config['path'])
                obj = getattr(module, config['object_name'])
                cls._objects[key] = obj

            except Exception as error:
                raise ModuleObjectLoaderError(
                    "Error loading module '{}.{}'.\nError Class: {}. Error Message: {}".format(
                        config['path'], config['object_name'],
                        error.__class__.__name__, str(error)
                    )
                )

        return obj


def build_item_key(name, *args):
    name = name.lower().replace(' ', '_')
    for arg in args:
        name += '_{}'.format(arg)
    return name


def get_items_model(item_type, store_id):
    item_types_model = get_model('item_types')
    return item_types_model.get_store_items_model(item_type, store_id)


def build_class_name(*names):
    final_name = ''
    for name in names:
        name = name.split(' ')
        for in_name in name:
            final_name += in_name.capitalize()

    return final_name + 'Model'


def extend_swagger_json(original, current_filename, swagger_json_name=None, by_method=False):
    swagger_json = deepcopy(original)
    additional_swagger = get_swagger_json(current_filename, swagger_json_name)

    if by_method:
        for path_name, addit_path in additional_swagger['paths'].items():
            if path_name in swagger_json['paths']:
                path = swagger_json['paths'][path_name]

                for method_name, method in addit_path.items():
                    if method_name in path:
                        path[method_name].update(method)
                    else:
                        path[method_name] = method

    else:
        swagger_json['paths'].update(additional_swagger['paths'])

    definitions = swagger_json.get('definitions')
    additional_definitions = additional_swagger.get('definitions')

    if additional_definitions:
        if definitions:
            definitions.update(additional_definitions)
        else:
            swagger_json['definitions'] = additional_definitions

    return swagger_json


def build_engine_object_key(engine_object):
    return '{}_{}_{}'.format(
        engine_object['strategy']['name'],
        engine_object['type'],
        engine_object['id']
    )


def makedirs(dir_):
    try:
        os.makedirs(dir_)
    except OSError as e:
        if os.errno.EEXIST != e.errno:
            raise


def run_coro(coro, session):
    if not asyncio.iscoroutine(coro):
        coro = _convert_future_to_coro(coro)

    if session.loop.is_running():
        return asyncio.run_coroutine_threadsafe(coro, session.loop).result()
    else:
        return session.loop.run_until_complete(coro)

async def _convert_future_to_coro(fut):
    return await fut



# MIT License

# Copyright (c) 2016 Diogo Dutra <dutradda@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from myreco.authorizer import MyrecoAuthorizer
from myreco.factory import ModelsFactory
from swaggerit.constants import SWAGGER_VALIDATOR, SWAGGER_SCHEMA
from swaggerit.aiohttp_api import AioHttpAPI
from swaggerit.exceptions import SwaggerItAPIError
from swaggerit.utils import get_swagger_json
from swaggerit.response import SwaggerResponse
from copy import deepcopy
import ujson


class MyrecoAPI(AioHttpAPI):
    def __init__(self, data_path, *, type_='recommender', sqlalchemy_bind=None, redis_bind=None,
                 elsearch_bind=None, swagger_json_template=None, title=None, version='1.0.0',
                 get_swagger_req_auth=True, loop=None, debug=False, factory_class=ModelsFactory,
                 extra_models=None):
        self.models_factory = factory_class(data_path)
        self.all_models = self.models_factory.make_all_models(type_)
        authorizer = MyrecoAuthorizer(self.all_models['users'])
        models = [model for model in self.all_models.values() if hasattr(model, '__api__')]

        if extra_models is not None:
            models.extend(extra_models)

        AioHttpAPI.__init__(
            self, models,
            sqlalchemy_bind=sqlalchemy_bind,
            redis_bind=redis_bind,
            elsearch_bind=elsearch_bind,
            swagger_json_template=swagger_json_template,
            title=title, version=version,
            authorizer=authorizer,
            get_swagger_req_auth=get_swagger_req_auth,
            loop=loop, debug=debug
        )
        self._set_items_metaschema_route('/doc')
        self._set_swagger_schema_route('/doc')

    def _set_items_metaschema_route(self, swagger_doc_url):
        self.items_metaschema = \
            get_swagger_json(__file__, 'item_types/store_items_metaschema.json')
        path = '/doc/items_metaschema.json'
        handler = self._set_handler_decorator(self._get_items_metaschema)
        self._set_route(path, 'GET', handler)

    async def _get_items_metaschema(self, req, session):
        deny = await self._authorize(req, session)
        if deny is not None:
            return deny

        headers = {'content-type': 'application/json'}
        return SwaggerResponse(200, headers, ujson.dumps(self.items_metaschema))

    def _set_swagger_schema_route(self, swagger_doc_url):
        handler = self._set_handler_decorator(self._get_swagger_schema)
        self._set_route('/doc/swagger_schema.json', 'GET', handler)

    async def _get_swagger_schema(self, req, session):
        deny = await self._authorize(req, session)
        if deny is not None:
            return deny

        headers = {'content-type': 'application/json'}
        return SwaggerResponse(200, headers, ujson.dumps(SWAGGER_SCHEMA))



# MIT License

# Copyright (c) 2016 Diogo Dutra <dutradda@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from myreco.engine_objects.object_base import EngineObjectBase
from myreco.exceptions import EngineError
import numpy as np
import ujson


class TopSellerArray(EngineObjectBase):

    def export(self, items_model, session):
        self._logger.info("Started export objects")

        readers = self._build_csv_readers()

        items_indices_map_dict = self._run_coro(
            self._get_items_indices_map_dict(items_model.indices_map, session),
            session
        )

        ret = self.update(readers, session, items_indices_map_dict)

        self._logger.info("Finished export objects")
        return ret

    def update(self, readers, session, items_indices_map_dict):
        self._build_top_seller_vector(readers, session, items_indices_map_dict)

        self._run_coro(
            session.redis_bind.set(
                self._redis_key,
                self._pack_array(self.numpy_array, compress=False)
            ),
            session
        )

        return {
            'length': int(self.numpy_array.size),
            'max_sells': int(max(self.numpy_array)),
            'min_sells': int(min(self.numpy_array))
        }

    def _build_top_seller_vector(self, readers, session, items_indices_map_dict):
        error_message = "No data found for engine object '{}'".format(self._engine_object['name'])
        if not len(readers):
            raise EngineError(error_message)

        indices_values_map = dict()

        for reader in readers:
            self._set_indices_values_map(indices_values_map, items_indices_map_dict, reader)

        if not indices_values_map:
            raise EngineError(error_message)

        vector = np.zeros(len(items_indices_map_dict), dtype=np.int32)
        indices = np.array(list(indices_values_map.keys()), dtype=np.int32)
        vector[indices] = np.array(list(indices_values_map.values()), dtype=np.int32)
        self.numpy_array = vector

    def _set_indices_values_map(self, indices_values_map, items_indices_map_dict, reader):
        for line in reader:
            line = ujson.loads(line)
            index = items_indices_map_dict.get(line['item_key'])
            if index is not None:
                indices_values_map[index] = int(line['value'])

    async def get_numpy_array(self, session):
        items_vector = await session.redis_bind.get(self._redis_key)
        if items_vector is not None:
            return self._unpack_array(items_vector, np.int32, compress=False)



# MIT License

# Copyright (c) 2016 Diogo Dutra <dutradda@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class ItemsIndicesDict(dict):

    def __init__(self, items_indices_map, items_model):
        self.items_model = items_model
        dict.__init__(self, items_indices_map)

    def __len__(self):
        values = self.values()
        if values:
            return max(self.values())+1
        else:
            return 0

    def get(self, key, default=None):
        if isinstance(key, str):
            key = key.encode()
        v = dict.get(self, key, default)
        return v


class ItemsIndicesMap(object):

    def __init__(self, items_model):
        self.items_model = items_model
        self.key = items_model.__key__ + '_indices_map'
        self.indices_items_key = items_model.__key__ + '_items_map'
        self.length_key = items_model.__key__ + '_indices_length'

    async def get_all(self, session):
        items_indices_map = await session.redis_bind.hgetall(self.key)
        items_indices_map = dict([(k, int(v.decode())) for k, v in items_indices_map.items()])
        return ItemsIndicesDict(items_indices_map, self.items_model)

    async def get_indices_items_map(self, session):
        map_ = await session.redis_bind.hgetall(self.indices_items_key)
        return {int(k): v.decode() for k, v in map_.items()}

    async def update(self, session):
        items_indices_map = await self.get_all(session)
        indices_items_map = await self.get_indices_items_map(session)

        items = await self.items_model.get_all(session)
        items_keys = self._build_keys(items)

        new_keys = [key for key in items_keys if key not in items_indices_map]
        old_keys = set([key for key in items_keys if key in items_indices_map])
        keys_to_delete = set(items_indices_map.keys()).difference(old_keys)
        free_indices = [v for k, v in items_indices_map.items() if k in keys_to_delete]

        [items_indices_map.pop(k) for k in keys_to_delete]
        [indices_items_map.pop(i) for i in free_indices]

        free_indices_length = len(free_indices)

        for key, index in zip(new_keys[:free_indices_length], free_indices):
            items_indices_map[key] = index
            indices_items_map[index] = key

        counter = len(items_indices_map)

        for key in new_keys[free_indices_length:]:
            items_indices_map[key] = counter
            indices_items_map[counter] = key
            counter += 1

        if keys_to_delete:
            await session.redis_bind.hdel(self.key, *keys_to_delete)
            await session.redis_bind.hdel(self.indices_items_key, *free_indices)

        if items_indices_map:
            await session.redis_bind.hmset_dict(self.key, items_indices_map)
            await session.redis_bind.hmset_dict(self.indices_items_key, indices_items_map)
            await session.redis_bind.set(self.length_key, len(items_indices_map))

        return self._format_output(await self.get_all(session))

    def _build_keys(self, items):
        return set([self.items_model.get_instance_key(item) for item in items])

    def _format_output(self, output):
        maximum_index = max(output.values()) if len(output.keys()) else None
        ret = {'total_items': len(output.keys()), 'maximum_index': maximum_index}
        return ret

    async def get_items(self, indices, session):
        if indices:
            return [item for item in \
                await session.redis_bind.hmget(self.indices_items_key, *indices) if item is not None]
        else:
            return []

    async def get_indices(self, keys, session):
        return [int(index.decode()) for index in \
            await session.redis_bind.hmget(self.key, *keys) if index is not None]

    async def get_length(self, session):
        len_ = await session.redis_bind.get(self.length_key)
        return None if len_ is None else int(len_)



# MIT License

# Copyright (c) 2016 Diogo Dutra <dutradda@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from swaggerit.utils import set_logger
from myreco.utils import build_engine_object_key, makedirs, run_coro
from myreco.exceptions import EngineError
from abc import abstractmethod, ABCMeta
from glob import glob
from gzip import GzipFile
import os.path
import asyncio
import zlib
import numpy as np


class EngineObjectBase(metaclass=ABCMeta):

    def __init__(self, engine_object, data_path=None):
        self._engine_object = engine_object
        self._set_redis_key()
        set_logger(self, self._redis_key)
        if data_path is not None:
            self._set_data_path(data_path)

    def _set_redis_key(self):
        self._redis_key = build_engine_object_key(self._engine_object)

    def _set_data_path(self, data_path):
        self._data_path = os.path.join(data_path, self._redis_key)
        makedirs(self._data_path)

    def _pack_array(self, array, compress=True, level=-1):
        if compress:
            return zlib.compress(array.tobytes(), level)
        else:
            return array.tobytes()

    def _unpack_array(self, array, dtype, compress=True):
        if array is not None:
            if compress:
                return np.fromstring(zlib.decompress(array), dtype=dtype)
            else:
                return np.fromstring(array, dtype=dtype)
        else:
            return None

    @abstractmethod
    def export(self, items_model, session):
        pass

    def _build_csv_readers(self, pattern=''):
        readers = []
        pattern = os.path.join(self._data_path, '{}*.gz'.format(pattern))

        for filename in glob(pattern):
            file_ = GzipFile(filename, 'r')
            readers.append(file_)

        return readers

    async def _get_items_indices_map_dict(self, items_indices_map, session):
        items_indices_map_dict = await items_indices_map.get_all(session)

        if not items_indices_map_dict.values():
            raise EngineError(
                "The Indices Map for '{}' is empty. Please update these items"
                .format(self._engine_object['item_type']['name']))

        return items_indices_map_dict

    def _run_coro(self, coro, session):
        return run_coro(coro, session)
    
    @abstractmethod
    def get_data(self, items_model, session):
        pass

    def _log_get_data_started(self):
        self._logger.info("Started import data")

    def _log_get_data_finished(self):
        self._logger.info("Finished import data")



# MIT License

# Copyright (c) 2016 Diogo Dutra <dutradda@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from myreco.engine_strategies.top_seller.array import TopSellerArray
from myreco.engine_strategies.strategy_base import EngineStrategyBase


class TopSellerEngineStrategy(EngineStrategyBase):
    configuration_schema = {
        'type': 'object',
        'required': ['top_seller_array'],
        'additionalProperties': False,
        'properties': {
            'top_seller_array': {
                'type': 'object',
                'required': ['days_interval'],
                'additionalProperties': False,
                'properties': {
                    'days_interval': {'type': 'integer'}
                }
            }
        }
    }
    object_types = {'top_seller_array': TopSellerArray}

    async def _build_items_vector(self, session, items_model, **external_variables):
        return await self.objects['top_seller_array'].get_numpy_array(session)



# MIT License

# Copyright (c) 2016 Diogo Dutra <dutradda@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from swaggerit.utils import get_swagger_json
from swaggerit.exceptions import SwaggerItModelError
from sqlalchemy.ext.declarative import AbstractConcreteBase, declared_attr
import sqlalchemy as sa


class EnginesModelBase(AbstractConcreteBase):
    __tablename__ = 'engines'
    __swagger_json__ = get_swagger_json(__file__)
    _jobs = dict()
    __table_args__ = (sa.UniqueConstraint('name', 'store_id'),)

    id = sa.Column(sa.Integer, primary_key=True)
    name = sa.Column(sa.String(255), nullable=False)

    @declared_attr
    def strategy_id(cls):
        return sa.Column(sa.ForeignKey('engine_strategies.id'), nullable=False)

    @declared_attr
    def store_id(cls):
        return sa.Column(sa.ForeignKey('stores.id'), nullable=False)

    @declared_attr
    def item_type_id(cls):
        return sa.Column(sa.ForeignKey('item_types.id'), nullable=False)

    @declared_attr
    def item_type(cls):
        return sa.orm.relationship('ItemTypesModel')

    @declared_attr
    def store(cls):
        return sa.orm.relationship('StoresModel')

    @declared_attr
    def strategy(cls):
        return sa.orm.relationship('EngineStrategiesModel')

    @declared_attr
    def objects(cls):
        return sa.orm.relationship('EngineObjectsModel', uselist=True,
                                   secondary='engines_objects')

    @property
    def variables(self):
        return self.strategy_instance.get_variables()

    @property
    def strategy_instance(self):
        if not hasattr(self, '_strategy_instance'):
            self._strategy_instance = \
                type(self.strategy).get_instance(self._todict_when_new())

        return self._strategy_instance

    def _todict_when_new(self):
        objects = []
        self_dict = self.todict({'variables': False})
        strategy_dict = self_dict['strategy']

        for obj in self.objects:
            obj_dict = obj.todict()
            obj_dict['strategy'] = strategy_dict
            objects.append(obj_dict)

        self_dict['objects'] = objects
        return self_dict

    async def _setattr(self, attr_name, value, session, input_):
        if attr_name == 'item_type_id':
            value = {'id': value}
            attr_name = 'item_type'

        if attr_name == 'strategy_id':
            value = {'id': value}
            attr_name = 'strategy'

        if attr_name == 'store_id':
            value = {'id': value}
            attr_name = 'store'

        await super()._setattr(attr_name, value, session, input_)

    async def _validate(self, session, input_):
        strategy_class = self.strategy.get_class()

        for obj in self.objects:
            if obj.type not in strategy_class.object_types:
                raise SwaggerItModelError(
                    "Invalid object type '{}'".format(obj.type),
                    instance=input_
                )

        props_sequence = [
            ('strategy_id', self.strategy.id),
            ('item_type_id',self.item_type.id),
            ('store_id', self.store.id)
        ]

        for obj in self.objects:
            for obj_prop_name, self_prop_value in props_sequence:
                value = getattr(obj, obj_prop_name)

                if value is None:
                    setattr(obj, obj_prop_name, self_prop_value)

                elif value != self_prop_value:
                    raise SwaggerItModelError(
                    "Invalid object '{}' with value '{}'. "\
                    "This value must be the same as '{}'".format(
                        obj_prop_name,
                        value,
                        self_prop_value
                    ),
                    instance=input_
                )

        self.strategy_instance.validate_config()

    async def init(self, session, input_=None, **kwargs):
        for object_ in kwargs.get('objects', []):
            object_['item_type_id'] = kwargs.get('item_type_id', self.item_type_id)
            object_['strategy_id'] = kwargs.get('strategy_id', self.strategy_id)
            object_['store_id'] = kwargs.get('store_id', self.store_id)

        await super().init(session, input_=input_, **kwargs)


    def _format_output_json(self, dict_inst, schema):
        if schema.get('variables') is not False:
            dict_inst['variables'] = self.variables


def build_engines_objects_table(metadata, **kwargs):
    return sa.Table(
        'engines_objects', metadata,
        sa.Column('engine_id', sa.Integer, sa.ForeignKey('engines.id', onupdate='CASCADE', ondelete='CASCADE'), primary_key=True),
        sa.Column('object_id', sa.Integer, sa.ForeignKey('engine_objects.id', onupdate='CASCADE', ondelete='CASCADE'), primary_key=True),
        **kwargs)



# MIT License

# Copyright (c) 2016 Diogo Dutra <dutradda@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from myreco.item_types.model import ItemTypesModelBase, ItemValidator
from myreco.utils import extend_swagger_json, run_coro
from swaggerit.exceptions import SwaggerItModelError
from tempfile import NamedTemporaryFile
from io import BytesIO
from collections import namedtuple
from zipfile import ZipFile
from gzip import GzipFile
import ujson
import boto3
import os
import asyncio
import gc


class ItemTypesDataFileImporterModelBase(ItemTypesModelBase):
    __swagger_json__ = extend_swagger_json(
        ItemTypesModelBase.__swagger_json__,
        __file__
    )

    @classmethod
    async def post_import_data_file_job(cls, req, session):
        content_type = req.headers.get('content-type')
        if content_type != 'application/zip' and content_type != 'application/gzip':
            raise SwaggerItModelError("Invalid content type '{}'".format(content_type))

        stream = BytesIO()
        async for line in req.body:
            stream.write(line)
        stream.seek(0)

        return await cls._create_job(
            cls._run_import_data_file_job,
            req, session, '_importer',
            stream=stream, content_type=content_type)

    @classmethod
    async def _create_job(cls, func, req, session, jobs_id_prefix, stream=None, content_type=None):
        store_id = req.query['store_id']
        store_items_model = await cls._get_store_items_model(req, session)
        if store_items_model is None:
            return cls._build_response(404)

        session = cls._copy_session(session)
        jobs_id = store_items_model.__key__ + jobs_id_prefix
        return super()._create_job(
            func, jobs_id,
            req, session,
            store_items_model, store_id,
            stream=stream, content_type=content_type
        )

    @classmethod
    def _run_import_data_file_job(
            cls, req, session, store_items_model,
            store_id, stream, content_type):
        upload_file = req.query.get('upload_file', True)
        if upload_file:
            cls._put_file_on_s3(stream, store_items_model, session, store_id)
            stream.seek(0)

        result = cls._update_items_from_zipped_file(stream, store_items_model, content_type, session)

        gc.collect()
        return result

    @classmethod
    def _put_file_on_s3(cls, stream, store_items_model, session, store_id):
        cls._logger.info("Started put file on S3 for '{}'".format(store_items_model.__key__))

        store = cls._run_coro(
            cls.get_model('stores').get(session, [{'id': store_id}]),
            session
        )[0]

        s3_bucket = store['configuration']['aws']['s3']['bucket']
        access_key_id = store['configuration']['aws'].get('access_key_id')
        secret_access_key = store['configuration']['aws'].get('secret_access_key')
        s3_key = '{}.zip'.format(store_items_model.__key__)

        boto3.resource(
            's3',
            aws_access_key_id=access_key_id,
            aws_secret_access_key=secret_access_key
        ).Bucket(s3_bucket).put_object(Body=stream, Key=s3_key)

        cls._logger.info("Finished put file on S3 for '{}'".format(store_items_model.__key__))

    @classmethod
    def _run_coro(cls, coro, session):
        return run_coro(coro, session)

    @classmethod
    def _update_items_from_zipped_file(cls, stream, store_items_model, content_type, session):
        feed = cls._unzip_file(content_type, stream)
        result = cls._update_items_from_file(feed, store_items_model, session)

        feed.close()

        if os.path.isfile(feed.name):
            os.remove(feed.name)

        return result

    @classmethod
    def _unzip_file(cls, content_type, stream=None, mode=None):
        tempfile = NamedTemporaryFile(delete=False)
        filename = tempfile.name
        kwargs = {'mode': mode} if mode else {}
        if stream is not None:
            tempfile.write(stream.read())

        tempfile.close()

        if content_type.endswith('gzip'):
            zfile = GzipFile(filename, **kwargs)

        else:
            zfile = ZipFile(filename, **kwargs)
            zfile = zfile.open(zfile.namelist()[0])

        return zfile


    @classmethod
    def _update_items_from_file(cls, feed, store_items_model, session):
        cls._logger.info(
            "Started update items from file for '{}'".format(store_items_model.__key__)
        )
        warning_message = "Invalid line for model '{}': ".format(store_items_model.__key__) + '{}'
        validator = ItemValidator(store_items_model.item_type['schema'])
        new_keys = set()
        success_lines = 0
        errors_lines = 0
        empty_lines = 0
        old_keys = cls._run_coro(
            session.redis_bind.hkeys(store_items_model.__key__),
            session
        )
        old_keys = set(old_keys)
        lines = []

        if hasattr(store_items_model, 'pre_process_feed'):
            feed = store_items_model.pre_process_feed(feed, session)

        for line in feed:
            try:
                line = cls._try_to_process_line(line, validator, store_items_model)
                if line is None:
                    empty_lines += 1
                    continue
                else:
                    success_lines += 1
                    new_keys.add(store_items_model.get_instance_key(line))
                    lines.append(line)

                    if len(lines) == 1000:
                        cls._run_coro(
                            store_items_model.insert(session, lines, skip_validation=True),
                            session
                        )
                        lines = []

            except Exception as error:
                errors_lines += 1
                cls._logger.warning(warning_message.format(line))
                cls._logger.warning(error)
                continue

        if lines:
            cls._run_coro(
                store_items_model.insert(session, lines, skip_validation=True),
                session
            )

        del lines
        old_keys.difference_update(new_keys)

        if old_keys:
            cls._run_coro(
                session.redis_bind.hdel(store_items_model.__key__, *old_keys),
                session
            )

        cls._run_coro(
            cls._set_stock_filter(store_items_model, session),
            session
        )

        cls._logger.info(
            "Finished update items from file for '{}'".format(store_items_model.__key__)
        )

        return {
            'success_lines': success_lines,
            'errors_lines': errors_lines,
            'empty_lines': empty_lines
        }

    @classmethod
    def _try_to_process_line(cls, line, validator, store_items_model):
        line = line.strip()

        if isinstance(line, bytes):
            line = line.decode()

        if not line:
            return None

        line = ujson.loads(line)
        validator.validate(line)

        return line

    @classmethod
    async def get_import_data_file_job(cls, req, session):
        return await cls._get_job('_importer', req, session)

    @classmethod
    async def _get_job(cls, sufix, req, session):
        store_items_model = await cls._get_store_items_model(req, session)
        if store_items_model is None:
            return cls._build_response(404)

        jobs_id = store_items_model.__key__ + sufix
        return await super()._get_job(jobs_id, req, session)



# MIT License

# Copyright (c) 2016 Diogo Dutra <dutradda@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from myreco.users.models import (GrantsModelBase, URIsModelBase, MethodsModelBase,
    UsersModelBase, build_users_grants_table, build_users_stores_table)
from myreco.stores.model import StoresModelBase
from myreco.external_variables.model import ExternalVariablesModelBase
from myreco.placements.models import (PlacementsModelBase, VariationsModelBase,
    ABTestUsersModelBase, build_variations_slots_table)
from myreco.slots.models import (SlotVariablesModelBase, SlotFiltersModelBase,
    SlotsModelBase, build_slots_fallbacks_table)
from myreco.engines.model import EnginesModelBase, build_engines_objects_table
from myreco.engine_objects.model import EngineObjectsModelBase
from myreco.engine_objects.data_importer.model import EngineObjectsDataImporterModelBase
from myreco.engine_objects.exporter.model import EngineObjectsExporterModelBase
from myreco.engine_strategies.model import EngineStrategiesModelBase
from myreco.item_types.model import ItemTypesModelBase, build_item_types_stores_table
from myreco.item_types.data_file_importer.model import ItemTypesDataFileImporterModelBase
from myreco.item_types.filters_updater.model import ItemTypesFiltersUpdaterModelBase
from swaggerit.models.orm.factory import FactoryOrmModels


class FactoryError(Exception):
    pass


class ModelsFactory(object):
    engine_strategies_base = EngineStrategiesModelBase
    engine_objects_recommender_base = EngineObjectsModelBase
    engine_objects_data_importer_base = EngineObjectsDataImporterModelBase
    engine_objects_exporter_base = EngineObjectsExporterModelBase
    engines_base = EnginesModelBase
    slots_base = SlotsModelBase
    slot_variables_base = SlotVariablesModelBase
    slot_filters_base = SlotFiltersModelBase
    item_types_recommender_base = ItemTypesModelBase
    item_types_data_importer_base = ItemTypesDataFileImporterModelBase
    item_types_objects_exporter_base = ItemTypesFiltersUpdaterModelBase
    placements_base = PlacementsModelBase
    variations_base = VariationsModelBase
    ab_test_users_base = ABTestUsersModelBase
    stores_base = StoresModelBase
    uris_base = URIsModelBase
    users_base = UsersModelBase
    methods_base = MethodsModelBase
    grants_base = GrantsModelBase
    external_variables_base = ExternalVariablesModelBase

    def __init__(self, data_path,
                 commons_models_attributes=None,
                 commons_tables_attributes=None):
        self.base_model = FactoryOrmModels.make_sqlalchemy_redis_base()
        self.base_model.__data_path__ = data_path
        self.meta_class = type(self.base_model)
        self._commons_models_attrs = self._init_attributes(commons_models_attributes)
        self._commons_tables_attrs = self._init_attributes(commons_tables_attributes)

    def _init_attributes(self, attrs=None, update=None):
        if attrs is None:
            attrs = dict()

        if update is not None:
            attrs.update(update)

        return attrs

    def make_all_models(self, app_type='recommender'):
        self.make_all_tables()
        app_types = {'recommender', 'data_importer', 'objects_exporter'}
        if (app_type not in app_types):
            raise FactoryError(
                "Invalid application type '{}'. Valid types: {}".format(
                    app_type, ', '.join(app_types)))

        return {
            'engine_strategies': self.make_engine_strategies_model(),
            'engine_objects': self.make_engine_objects_model(app_type),
            'engines': self.make_engines_model(),
            'slots': self.make_slots_model(),
            'slot_variables': self.make_slot_variables_model(),
            'slot_filters': self.make_slot_filters_model(),
            'item_types': self.make_item_types_model(app_type),
            'placements': self.make_placements_model(),
            'variations': self.make_variations_model(),
            'ab_test_users': self.make_ab_test_users_model(),
            'stores': self.make_stores_model(),
            'uris': self.make_uris_model(),
            'users': self.make_users_model(),
            'methods': self.make_methods_model(),
            'grants': self.make_grants_model(),
            'external_variables': self.make_external_variables_model()
        }

    def make_all_tables(self):
        return {
            'users_grants': self.make_users_grants_table(),
            'users_stores': self.make_users_stores_table(),
            'variations_slots': self.make_variations_slots_table(),
            'slots_fallbacks': self.make_slots_fallbacks_table(),
            'item_types_stores': self.make_item_types_stores_table(),
            'engines_objects': self.make_engines_objects_table()
        }

    def make_users_grants_table(self, attributes=None):
        attributes = self._init_attributes(attributes, self._commons_tables_attrs)
        return build_users_grants_table(self.base_model.metadata, **attributes)

    def make_users_stores_table(self, attributes=None):
        attributes = self._init_attributes(attributes, self._commons_tables_attrs)
        return build_users_stores_table(self.base_model.metadata, **attributes)

    def make_variations_slots_table(self, attributes=None):
        attributes = self._init_attributes(attributes, self._commons_tables_attrs)
        return build_variations_slots_table(self.base_model.metadata, **attributes)

    def make_slots_fallbacks_table(self, attributes=None):
        attributes = self._init_attributes(attributes, self._commons_tables_attrs)
        return build_slots_fallbacks_table(self.base_model.metadata, **attributes)

    def make_item_types_stores_table(self, attributes=None):
        attributes = self._init_attributes(attributes, self._commons_tables_attrs)
        return build_item_types_stores_table(self.base_model.metadata, **attributes)

    def make_engines_objects_table(self, attributes=None):
        attributes = self._init_attributes(attributes, self._commons_tables_attrs)
        return build_engines_objects_table(self.base_model.metadata, **attributes)

    def make_engine_strategies_model(self, base=EngineStrategiesModelBase, attributes=None):
        attributes = self._init_attributes(attributes, self._commons_models_attrs)
        return self.meta_class('EngineStrategiesModel', (base, self.base_model), attributes)

    def make_engine_objects_model(self, app_type, attributes=None):
        recommender_base = type(self).engine_objects_recommender_base
        data_importer_base = type(self).engine_objects_data_importer_base
        objects_exporter_base = type(self).engine_objects_exporter_base
        attributes = self._init_attributes(attributes, self._commons_models_attrs)

        if app_type == 'recommender':
            bases_classes = (recommender_base, self.base_model)
        elif app_type == 'data_importer':
            bases_classes = (data_importer_base, self.base_model)
        elif app_type == 'objects_exporter':
            bases_classes = (objects_exporter_base, self.base_model)

        return self.meta_class('EngineObjectsModel', bases_classes, attributes)

    def make_engines_model(self, attributes=None):
        base = type(self).engines_base
        attributes = self._init_attributes(attributes, self._commons_models_attrs)
        return self.meta_class('EnginesModel', (base, self.base_model), attributes)

    def make_slots_model(self, attributes=None):
        base = type(self).slots_base
        attributes = self._init_attributes(attributes, self._commons_models_attrs)
        return self.meta_class('SlotsModel', (base, self.base_model), attributes)

    def make_slot_variables_model(self, attributes=None):
        base = type(self).slot_variables_base
        attributes = self._init_attributes(attributes, self._commons_models_attrs)
        return self.meta_class('SlotVariablesModel', (base, self.base_model), attributes)

    def make_slot_filters_model(self, attributes=None):
        base = type(self).slot_filters_base
        attributes = self._init_attributes(attributes, self._commons_models_attrs)
        return self.meta_class('SlotFiltersModel', (base, self.base_model), attributes)

    def make_item_types_model(self, app_type, attributes=None):
        recommender_base = type(self).item_types_recommender_base
        data_importer_base = type(self).item_types_data_importer_base
        objects_exporter_base = type(self).item_types_objects_exporter_base
        attributes = self._init_attributes(attributes, self._commons_models_attrs)

        if app_type == 'recommender':
            bases_classes = (recommender_base, self.base_model)
        elif app_type == 'data_importer':
            bases_classes = (data_importer_base, self.base_model)
        elif app_type == 'objects_exporter':
            bases_classes = (objects_exporter_base, self.base_model)

        return self.meta_class('ItemTypesModel', bases_classes, attributes)

    def make_placements_model(self, attributes=None):
        base = type(self).placements_base
        attributes = self._init_attributes(attributes, self._commons_models_attrs)
        return self.meta_class(
            'PlacementsModel', (base, self.base_model), attributes)

    def make_variations_model(self, attributes=None):
        base = type(self).variations_base
        attributes = self._init_attributes(attributes, self._commons_models_attrs)
        return self.meta_class(
            'VariationsModel', (base, self.base_model), attributes)

    def make_ab_test_users_model(self, attributes=None):
        base = type(self).ab_test_users_base
        attributes = self._init_attributes(attributes, self._commons_models_attrs)
        return self.meta_class(
            'ABTestUsersModel', (base, self.base_model), attributes)

    def make_stores_model(self, attributes=None):
        base = type(self).stores_base
        attributes = self._init_attributes(attributes, self._commons_models_attrs)
        return self.meta_class(
            'StoresModel', (base, self.base_model), attributes)

    def make_uris_model(self, attributes=None):
        base = type(self).uris_base
        attributes = self._init_attributes(attributes, self._commons_models_attrs)
        return self.meta_class(
            'URIsModel', (base, self.base_model), attributes)

    def make_users_model(self, attributes=None):
        base = type(self).users_base
        attributes = self._init_attributes(attributes, self._commons_models_attrs)
        return self.meta_class('UsersModel', (base, self.base_model), attributes)

    def make_methods_model(self, attributes=None):
        base = type(self).methods_base
        attributes = self._init_attributes(attributes, self._commons_models_attrs)
        return self.meta_class('MethodsModel', (base, self.base_model), attributes)

    def make_grants_model(self, attributes=None):
        base = type(self).grants_base
        attributes = self._init_attributes(attributes, self._commons_models_attrs)
        return self.meta_class('GrantsModel', (base, self.base_model), attributes)

    def make_external_variables_model(self, attributes=None):
        base = type(self).external_variables_base
        attributes = self._init_attributes(attributes, self._commons_models_attrs)
        return self.meta_class('ExternalVariablesModel', (base, self.base_model), attributes)



# MIT License

# Copyright (c) 2017 Diogo Dutra <dutradda@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from sqlalchemy.ext.declarative import AbstractConcreteBase
from myreco.utils import ModuleObjectLoader
from swaggerit.utils import get_swagger_json
import sqlalchemy as sa
import ujson


class EngineStrategiesModelBase(AbstractConcreteBase):
    __tablename__ = 'engine_strategies'
    __swagger_json__ = get_swagger_json(__file__)
    _jobs = dict()
    __table_args__ = (sa.UniqueConstraint('class_name', 'class_module'),)

    id = sa.Column(sa.Integer, primary_key=True)
    name = sa.Column(sa.String(255), unique=True, nullable=False)
    class_name = sa.Column(sa.String(255), nullable=False)
    class_module = sa.Column(sa.String(255), nullable=False)

    async def _validate(self, session, input_):
        self.get_class()

    @classmethod
    def _get_class(cls, strategy):
        return ModuleObjectLoader.load({
            'path': strategy['class_module'],
            'object_name': strategy['class_name']
        })

    def get_class(self):
        if not hasattr(self, '_class'):
            self._class = type(self)._get_class(self.todict({'object_types': False}))

        return self._class

    @property
    def object_types(self):
        return self.get_class().object_types.keys()

    def _format_output_json(self, dict_inst, schema):
        if schema.get('object_types') is not False:
            dict_inst['object_types'] = self.object_types

    @classmethod
    def get_instance(cls, engine, items_model=None):
        return cls._get_class(engine['strategy'])(engine, items_model)



# MIT License

# Copyright (c) 2016 Diogo Dutra <dutradda@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from myreco.engine_strategies.filters.filters import BooleanFilterBy
from myreco.engine_strategies.filters.factory import FiltersFactory
from myreco.item_types.data_file_importer.model import ItemTypesDataFileImporterModelBase
from myreco.utils import extend_swagger_json
from swaggerit.exceptions import SwaggerItModelError


class ItemTypesFiltersUpdaterModelBase(ItemTypesDataFileImporterModelBase):
    __swagger_json__ = extend_swagger_json(
        ItemTypesDataFileImporterModelBase.__swagger_json__,
        __file__
    )

    @classmethod
    async def post_update_filters_job(cls, req, session):
        return await cls._create_job(cls._run_update_filters_job, req, session, '_updater')

    @classmethod
    async def _run_update_filters_job(cls, req, session, store_items_model, store_id, **kwargs):
        cls._logger.info("Started update filters for '{}'".format(store_items_model.__key__))

        return await cls.update_enabled_filters(store_items_model, session, store_id)

    @classmethod
    async def get_update_filters_job(cls, req, session):
        return await cls._get_job('_updater', req, session)

    @classmethod
    async def update_enabled_filters(cls, store_items_model, session, store_id):
        items_indices_map = store_items_model.indices_map
        items_indices_map_ret = await items_indices_map.update(session)
        items_indices_map_len = await items_indices_map.get_length(session)

        filters_factory = cls.get_model('slot_filters').__factory__
        enabled_filters = await cls._get_enabled_filters(store_items_model, session, store_id)
        filters_ret = dict()
        items_indices_map_dict = await items_indices_map.get_all(session)
        items = await cls._get_items_with_indices_and_stock(
            store_items_model, session, items_indices_map_dict
        )

        stock_filter = BooleanFilterBy(store_items_model, 'stock')
        await stock_filter.update(session, items, items_indices_map_len)

        for slot_filter, schema in enabled_filters:
            filter_ = filters_factory.make(
                store_items_model, slot_filter,
                schema, slot_filter['skip_values']
            )
            filters_ret[filter_.name] = await filter_.update(session, items, items_indices_map_len)

        cls._logger.info("Finished update filters for '{}'".format(store_items_model.__key__))
        return {'items_indices_map': items_indices_map_ret, 'filters': filters_ret}

    @classmethod
    async def _get_enabled_filters(cls, store_items_model, session, store_id):
        slots_model = cls.get_model('slots')
        slots = await slots_model.get(session, **{'store_id': store_id})
        filters_external_variables = []

        # used to aggregate filters inclusive and exclusive with same property
        filters_names_set = set()

        for slot in slots:
            if store_items_model.item_type['id'] == slot['engine']['item_type_id']:
                for slot_filter in slot['slot_filters']:
                    filter_name = slot_filter['property_name']
                    if filter_name not in filters_names_set:
                        schema = \
                            store_items_model.item_type['schema']['properties'][filter_name]
                        filters_external_variables.append((slot_filter, schema))
                        filters_names_set.add(filter_name)

        return filters_external_variables

    @classmethod
    async def _get_items_with_indices_and_stock(cls, store_items_model, session, items_indices_map_dict):
        items = []
        page = 1
        items_part = await store_items_model.get(session, page=page, items_per_page=100000)

        while items_part:
            items.extend(items_part)
            page += 1
            items_part = await store_items_model.get(session, page=page, items_per_page=100000)

        for item in items:
            item_key = store_items_model.get_instance_key(item)
            index = items_indices_map_dict.get(item_key)
            if index is not None:
                item['index'] = index
                item['stock'] = True

        return items



# MIT License

# Copyright (c) 2017 Diogo Dutra <dutradda@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from swaggerit.utils import get_swagger_json
from swaggerit.exceptions import SwaggerItModelError
from sqlalchemy.ext.declarative import AbstractConcreteBase, declared_attr
import sqlalchemy as sa
import ujson
import jsonschema


class EngineObjectsModelBase(AbstractConcreteBase):
    __tablename__ = 'engine_objects'
    __swagger_json__ = get_swagger_json(__file__)
    __table_args__ = (sa.UniqueConstraint('name', 'type', 'store_id'),)

    id = sa.Column(sa.Integer, primary_key=True)
    type = sa.Column(sa.String(255), nullable=False)
    name = sa.Column(sa.String(255), nullable=False)
    configuration_json = sa.Column(sa.Text, nullable=False)

    @declared_attr
    def strategy_id(cls):
        return sa.Column(sa.ForeignKey('engine_strategies.id'), nullable=False)

    @declared_attr
    def store_id(cls):
        return sa.Column(sa.ForeignKey('stores.id'), nullable=False)

    @declared_attr
    def item_type_id(cls):
        return sa.Column(sa.ForeignKey('item_types.id'), nullable=False)

    @declared_attr
    def strategy(cls):
        return sa.orm.relationship('EngineStrategiesModel')

    @declared_attr
    def item_type(cls):
        return sa.orm.relationship('ItemTypesModel')

    @declared_attr
    def store(cls):
        return sa.orm.relationship('StoresModel')

    @property
    def configuration(self):
        if not hasattr(self, '_configuration'):
            self._configuration = ujson.loads(self.configuration_json)
        return self._configuration

    async def _setattr(self, attr_name, value, session, input_):
        if attr_name == 'configuration':
            value = ujson.dumps(value)
            attr_name = 'configuration_json'

        if attr_name == 'strategy_id':
            value = {'id': value}
            attr_name = 'strategy'

        await super()._setattr(attr_name, value, session, input_)

    def _format_output_json(self, dict_inst, schema):
        if schema.get('configuration') is not False:
            dict_inst.pop('configuration_json')
            dict_inst['configuration'] = self.configuration

    async def _validate(self, session, input_):
        # disable validation when the operation was did by the engine model
        if (isinstance(input_, list) and 'objects' in input_[0]) or \
                (isinstance(input_, dict) and 'objects' in input_):
            return

        strategy_class = self.strategy.get_class()

        if self.type not in strategy_class.object_types:
            raise SwaggerItModelError(
                "Invalid object type '{}'".format(self.type),
                instance=input_
            )

        object_schema = strategy_class.configuration_schema['properties'][self.type]

        if 'definitions' in strategy_class.configuration_schema:
            object_schema['definitions'] = strategy_class.configuration_schema['definitions']

        jsonschema.validate(self.configuration, object_schema)



# MIT License

# Copyright (c) 2016 Diogo Dutra <dutradda@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from myreco.item_types._store_items_model_meta import _StoreItemsModelBaseMeta
from myreco.utils import build_item_key, build_class_name, ModuleObjectLoader
from myreco.engine_strategies.filters.filters import BooleanFilterBy
from swaggerit.utils import get_swagger_json, get_dir_path
from swaggerit.method import SwaggerMethod
from swaggerit.request import SwaggerRequest
from swaggerit.models.orm.factory import FactoryOrmModels
from sqlalchemy.ext.declarative import declared_attr, AbstractConcreteBase
from jsonschema import ValidationError, validate
from jsonschema.validators import create, Draft4Validator
from copy import deepcopy
import sqlalchemy as sa
import ujson


store_items_metaschema = get_swagger_json(__file__, 'store_items_metaschema.json')
ItemValidator = create(store_items_metaschema, Draft4Validator.VALIDATORS)
ItemValidator.DEFAULT_TYPES['simpleObject'] = dict


class _ItemTypesModelBase(AbstractConcreteBase):
    __tablename__ = 'item_types'
    __swagger_json__ = get_swagger_json(__file__)
    __schema_dir__ = get_dir_path(__file__)

    id = sa.Column(sa.Integer, primary_key=True)
    name = sa.Column(sa.String(255), unique=True, nullable=False)
    schema_json = sa.Column(sa.Text, nullable=False)
    store_items_class_json = sa.Column(sa.Text)

    @declared_attr
    def stores(cls):
        return sa.orm.relationship('StoresModel', uselist=True, secondary='item_types_stores')

    @property
    def store_items_class(self):
        if not hasattr(self, '_store_items_class'):
            self._store_items_class = \
                ujson.loads(self.store_items_class_json) if self.store_items_class_json \
                    is not None else None
        return self._store_items_class

    async def _setattr(self, attr_name, value, session, input_):
        if attr_name == 'schema':
            self._validate_input(value)
            value = ujson.dumps(value)
            attr_name = 'schema_json'

        if attr_name == 'store_items_class':
            value = ujson.dumps(value)
            attr_name = 'store_items_class_json'

        await super()._setattr(attr_name, value, session, input_)

    def _validate_input(self, schema):
        validate(schema, get_swagger_json(__file__, 'store_items_metaschema.json'))

        for id_name in schema['id_names']:
            if id_name not in schema.get('properties', {}):
                raise ValidationError(
                    "id_name '{}' was not found in schema properties".format(id_name),
                    instance=schema['id_names'], schema=schema)

    def _format_output_json(self, dict_inst, todict_schema):
        if todict_schema.get('schema') is not False:
            if 'schema_json' in dict_inst:
                dict_inst['schema'] = ujson.loads(dict_inst.pop('schema_json'))

                schema_properties = dict_inst['schema'].get('properties', {})
                schema_properties_names = sorted(schema_properties.keys())
                dict_inst['available_filters'] = \
                    [{'name': name, 'schema': schema_properties[name]} \
                        for name in schema_properties_names if name != '_operation']

        if todict_schema.get('store_items_class') is not False:
            dict_inst.pop('store_items_class_json')
            dict_inst['store_items_class'] = self.store_items_class

    @classmethod
    async def swagger_get_filter_types(cls, req, session):
        filters_factory = cls.get_model('slot_filters').__factory__
        body = ujson.dumps(filters_factory.get_filter_types())
        return cls._build_response(200, body=body)


class _StoreItemsOperationsMixin(object):

    @classmethod
    async def swagger_insert_items(cls, req, session):
        store_items_model = await cls._get_store_items_model(req, session)
        if store_items_model is None:
            return cls._build_response(404)

        resp = await store_items_model.swagger_insert(req, session)
        await cls._set_stock_filter(store_items_model, session)
        return resp

    @classmethod
    async def _get_store_items_model(cls, req, session):
        id_ = req.path_params['id']
        store_id = req.query.pop('store_id')
        item_type = await cls._get_item_type(id_, store_id, session)
        if item_type is None:
            return None

        return cls.get_store_items_model(item_type, store_id)

    @classmethod
    async def _get_item_type(cls, id_, store_id, session):
        item_types = await cls.get(session, {'id': id_})
        if not item_types or not (item_types and cls._has_store(item_types[0], store_id)):
            return None

        return item_types[0]

    @classmethod
    def _has_store(cls, item_type, store_id):
        for store in item_type['stores']:
            if store['id'] == store_id:
                return True

        return False

    @classmethod
    def get_store_items_model(cls, item_type, store_id):
        store_items_model_key = build_item_key('store_items', item_type['name'], store_id)
        store_items_model = cls.get_model(store_items_model_key)

        if store_items_model is None:
            store_items_model = \
                cls._set_store_items_model(item_type, store_items_model_key, store_id)

        return store_items_model

    @classmethod
    def _set_store_items_model(cls, item_type, store_items_model_key, store_id):
        class_name = build_class_name(item_type['name'], str(store_id))
        base_class = cls._get_store_items_class(item_type)
        return FactoryOrmModels.make_redis_elsearch(
            class_name, item_type['schema']['id_names'],
            store_items_model_key, use_elsearch=False,
            metaclass=_StoreItemsModelBaseMeta,
            base=base_class,
            extra_attributes={
                'insert_validator': cls._build_insert_validator(item_type),
                'update_validator': cls._build_update_validator(item_type),
                'item_type': item_type
            }
        )

    @classmethod
    def _get_store_items_class(cls, item_type):
        return ModuleObjectLoader.load({
            'path': item_type['store_items_class']['module'],
            'object_name': item_type['store_items_class']['class_name']
        }) if item_type['store_items_class'] else object

    @classmethod
    def _build_insert_validator(cls, item_type):
        return ItemValidator({
            'type': 'array',
            'minItems': 1,
            'items': item_type['schema']
        })

    @classmethod
    def _build_update_validator(cls, item_type):
        schema = deepcopy(item_type['schema'])
        properties = schema.get('properties')
        if properties:
            properties['_operation'] = {'enum': ['delete', 'update']}

        return ItemValidator({
            'type': 'array',
            'minItems': 1,
            'items': schema
        })

    @classmethod
    async def swagger_update_items(cls, req, session):
        store_items_model = await cls._get_store_items_model(req, session)
        if store_items_model is None:
            return cls._build_response(404)

        resp = await store_items_model.swagger_update_many(req, session)
        await cls._set_stock_filter(store_items_model, session)
        return resp

    @classmethod
    async def swagger_get_item(cls, req, session):
        store_items_model = await cls._get_store_items_model(req, session)
        if store_items_model is None:
            return cls._build_response(404)

        req = SwaggerRequest(
            req.path, req.method,
            scheme=req.scheme,
            host=req.host,
            path_params=req.path_params['item_key'],
            query=req.query,
            headers=req.headers,
            body=req.body,
            body_schema=req.body_schema,
            context=req.context
        )
        return await store_items_model.swagger_get(req, session)

    @classmethod
    async def swagger_get_all_items(cls, req, session):
        store_items_model = await cls._get_store_items_model(req, session)
        if store_items_model is None:
            return cls._build_response(404)

        return await store_items_model.swagger_get_all(req, session)

    @classmethod
    async def swagger_search_items(cls, req, session):
        store_items_model = await cls._get_store_items_model(req, session)
        if store_items_model is None:
            return cls._build_response(404)

        return await store_items_model.swagger_search(req, session)

    @classmethod
    async def _set_stock_filter(cls, store_items_model, session):
        items_indices_map_dict = await store_items_model.indices_map.get_all(session)
        items_indices_map_len = await store_items_model.indices_map.get_length(session)

        if items_indices_map_dict.values():
            items_keys = set(await session.redis_bind.hkeys(store_items_model.__key__))
            items_indices_keys = set(items_indices_map_dict.keys())
            remaining_keys = items_indices_keys.intersection(items_keys)
            old_keys = items_indices_keys.difference(items_keys)

            items = []
            cls._set_stock_item(
                store_items_model, remaining_keys,
                items_indices_map_dict, True, items
            )
            cls._set_stock_item(store_items_model, old_keys, items_indices_map_dict, False, items)

            stock_filter = BooleanFilterBy(store_items_model, 'stock')
            await stock_filter.update(session, items, items_indices_map_len)

    @classmethod
    def _set_stock_item(cls, store_items_model, keys, items_indices_map_dict, value, items):
        for key in keys:
            item = {}
            store_items_model.set_instance_ids(item, key)
            item.update({'stock': value, 'index': int(items_indices_map_dict[key])})
            items.append(item)


class ItemTypesModelBase(_ItemTypesModelBase, _StoreItemsOperationsMixin):
    pass


def build_item_types_stores_table(metadata, **kwargs):
    return sa.Table(
        'item_types_stores', metadata,
        sa.Column('item_type_id', sa.Integer, sa.ForeignKey('item_types.id', ondelete='CASCADE', onupdate='CASCADE'), primary_key=True),
        sa.Column('store_id', sa.Integer, sa.ForeignKey('stores.id', ondelete='CASCADE', onupdate='CASCADE'), primary_key=True),
        **kwargs)
